import java.io.File;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.HashMap;
import java.util.Vector;

//Class FileManager: handle all input/output operations needed by Gasoline algorithm
public class FileManager
{
	//Read all graphs in the directory specified by a path and store them in internal objects
	//@path: path of the directory containing all input networks for the algorithm
	public Graph[] readGraphs(String path)
	{
		File f=new File(path);
		File[] list=f.listFiles();
		Graph[] g=new Graph[list.length];
		int i=0;
		for(i=0;i<list.length;i++)
		{
			//Create an empty graph whose name is the filename
			g[i]=new Graph(list[i].getName().toUpperCase());
			try
			{
				BufferedReader br=new BufferedReader(new FileReader(list[i]));
				br.readLine();
				String str="";
				while((str=br.readLine())!=null)
				{
					//Read an edge of the network (with optional associated weight)
					String[] split=str.split("\t");
					String idSource=split[0];
					String idDest=split[1];
					//double weight=Double.parseDouble(split[2]);
					g[i].addNode(idSource);
					g[i].addNode(idDest);
					//g[i].addArc(idSource,idDest,weight);
					g[i].addArc(idSource,idDest);
				}
				br.close();
			}
			catch(Exception e)
			{}
		}
		return g;
	}
	
	//Read the file with protein-protein homology scores (node label similarities)
	//@homologyFile: path of the homology file
	public HashMap<String,Double> readHomologyScores(String homologyFile)
	{
		int i=0;
		//Map of homology scores. The key is the couple of proteins and the value is their homology similarity
		//Homology similarity is the Blast E-Value 
		HashMap<String,Double> mapHomology=new HashMap<String,Double>();
		try
		{
			BufferedReader br=new BufferedReader(new FileReader(homologyFile));
			String str="";
			while((str=br.readLine())!=null)
			{
				String[] split=str.split("\t");
				String idSource=split[0];
				String idDest=split[1];
				double weight=Double.parseDouble(split[2]);
				//Convert the blast e-value score to the blast bit score
				if(weight==0.0)
					weight=2000;
				else
					weight=-(Math.log(weight)/Math.log(2));
				//Put the score of the couple irrespectively of the protein ids order
				mapHomology.put(idSource+"-"+idDest,weight);
				mapHomology.put(idDest+"-"+idSource,weight);
			}
			br.close();
		}
		catch(Exception e)
		{}
		return mapHomology;
	}
	
	//Write the final set of local alignments to different text files
	//@g: the set of input graphs
	//@rankAlign: the list of local alignments ordered by size and ISC score
	//@pathOutput: directory where output text files will be saved
	public static void writeAlignments(Graph[] g, OrderedList<Alignment> rankAlign, String pathOutput)
	{
		//Create output directory if it does not exist
		if(pathOutput!=null)
		{
			File f=new File(pathOutput);
			if(!f.exists())
				f.mkdir();
		}
		//Scan the list of alignments, starting from the one with highest score
		NodeOrdList<Alignment> aux=rankAlign.getMax();
		int cont=1;
		int i=0, j=0, k=0;
		while(aux!=null)
		{
			//Store the alignment into a text file
			Alignment a=aux.getInfo();
			//Mapping of the local alignment
			Vector<String>[] mapping=a.getMapping();
			if(pathOutput!=null)
			{
				try
				{
					BufferedWriter bw=new BufferedWriter(new FileWriter(pathOutput+"/Alignment"+cont+".txt"));
					bw.write("Alignment score: "+a.getAlignSize()*a.getIscScore()+"\r\n");
					bw.write("Alignment size: "+a.getAlignSize()+"\r\n");
					bw.write("Index of Structural Conservation (ISC): "+a.getIscScore()+"\r\n"+"\r\n");
					for(j=0;j<mapping.length;j++)
					{
						bw.write(g[j].getName()+":\r\n");
						bw.write(g[j].buildInducedSubgraph(mapping[j])+"\r\n");
					}
					bw.write("MAPPING:\r\n"+"\r\n");
					for(j=0;j<mapping[0].size();j++)
					{
						for(i=0;i<mapping.length-1;i++)
							bw.write(mapping[i].get(j)+"\t");
						bw.write(mapping[i].get(j));
						bw.write("\r\n");
					}
					bw.close();
				}
				catch(Exception e)
				{}
			}
			aux=aux.getNext();
			cont++;
		}
		//System.out.println();
	}
}