import java.util.HashMap;
import java.util.Vector;
import java.util.HashSet;
import java.util.Iterator;

//class Gasoline: contains all the code relative to Gasoline algorithm
//@pathNetworks: path of the directory where input networks are located
//@sigma: minimum degree of initial seed nodes
//@overlapThresh: maximum percentage of overlap between two distinct local alignments (between 0 and 1)
//@iterIterative: number of iteration of the iterative phase (extension+removal steps)
//@minAlignSize: minimum size, i.e. number of aligned nodes, of local alignments (>=2)
//@structConservationThresh: minimum value of structural conservation (ISC) score (between 0 and 1)
//@pathOutput: output directory where local alignments will be stored

public class Gasoline
{
	//Input parameters
	private String pathNetworks="Networks";
	private int sigma=3;
	private double overlapThresh=0.5;
	private int iterIterative=20;
	private int minAlignSize=4;
	private double structConservationThresh=0.8;
	private String pathOutput=null;
	public Gasoline(String pathNetworks, int sigma, double overlapThresh, int iterIterative, int minAlignSize, double structConservationThresh, String pathOutput)
	{
		this.pathNetworks=pathNetworks;
		this.sigma=sigma;
		this.overlapThresh=overlapThresh;
		this.iterIterative=iterIterative;
		this.minAlignSize=minAlignSize;
		this.structConservationThresh=structConservationThresh;
		this.pathOutput=pathOutput;
	}
	//Run Gasoline algorithm
	public OrderedList<Alignment> runGasoline()
	{
		//Initialize the list of local alignments found by Gasoline
		OrderedList<Alignment> rankAlign=new OrderedList<Alignment>();
		FileManager fileMan=new FileManager();
		int i=0, j=0, k=0;
		int indIter=0;
        
		//Read input networks
		Graph[] g=fileMan.readGraphs(pathNetworks);
		
		//Compute the number of required iteration for Gibbs sampling
		int iterGibbs=0;
		double prob=1.0;
		while(true)
		{
			prob=prob*((double)(g.length-1)/g.length);
			iterGibbs++;
			if(prob<0.001)
				break;
		}
		GibbsSampler gibbs=new GibbsSampler(iterGibbs);
		
		//Start algorithm
		long inizio=System.currentTimeMillis();
		
		//Filter starting nodes
		HashSet<String>[] startingNodes=filterStartingNodes(g, sigma);
		boolean search=true;
		
		//Start search for local alignments
		while(search)
		{
			//Check if all seeds have been analyzed
			for(i=0;i<startingNodes.length;i++)
			{
				if(startingNodes[i].isEmpty())
				{
					search=false;
					break;
				}
			}
			if(search)
			{
				//PHASE 1) Search for an optimal alignment of seeds
				String[][] candidates=new String[startingNodes.length][];
				for(i=0;i<startingNodes.length;i++)
				{
					candidates[i]=new String[startingNodes[i].size()];
					j=0;
					Iterator<String> it=startingNodes[i].iterator();
					while(it.hasNext())
					{
						String id=it.next();
						candidates[i][j]=id;
						j++;
					}
				}
				String[] alignNodes=gibbs.runGibbs(candidates,null);
			
				//Check if some aligned nodes has no homologs, and skip extension if necessary. Otherwise, save partial alignment and try to extend it
				Vector<String>[] finalAlign=new Vector[g.length];
					for(i=0;i<finalAlign.length;i++)
					{
						finalAlign[i]=new Vector<String>();
						finalAlign[i].add(alignNodes[i]);
					}
					//Save current alignment
					double scoreTopology=getScoreTopology(g,finalAlign);
					Vector<String>[] copyAlign=new Vector[g.length];
					for(i=0;i<finalAlign.length;i++)
					{
						copyAlign[i]=new Vector<String>();
						for(j=0;j<finalAlign[i].size();j++)
							copyAlign[i].add(finalAlign[i].get(j));
					}
					Alignment a=new Alignment(copyAlign, copyAlign[0].size(), scoreTopology);
			
					//PHASE 2) Iterative phase: removal steps followed by extension steps
					for(i=0;i<iterIterative;i++)
					{
						//If alignment has at least size 3, perform removal step
						if(finalAlign[0].size()>=3)
							removeNode(g,finalAlign,structConservationThresh);
						extendSeeds(g,finalAlign,iterGibbs,structConservationThresh);
						//Update best alignment if necessary
						if(finalAlign[0].size()>a.getAlignSize())
						{
							scoreTopology=getScoreTopology(g,finalAlign);
							copyAlign=new Vector[g.length];
							for(i=0;i<finalAlign.length;i++)
							{
								copyAlign[i]=new Vector<String>();
								for(j=0;j<finalAlign[i].size();j++)
									copyAlign[i].add(finalAlign[i].get(j));
							}
							a=new Alignment(copyAlign, copyAlign[0].size(), scoreTopology);
						}
					}
			
					//Print results and save alignment to final list, provided that the alignment size is above a threshold
					Vector<String>[] mapping=a.getMapping();
					if(mapping[0].size()>=minAlignSize)
					{
						rankAlign.insertOrdered(a);
						System.out.println("ALIGNMENT SIZE: "+a.getAlignSize());
						System.out.println("ISC_SCORE: "+a.getIscScore());    
						System.out.println("ITERATION: "+(indIter+1)+"\n");
					}
				//Remove starting seeds
				for(i=0;i<alignNodes.length;i++)
					startingNodes[i].remove(alignNodes[i]);
			}
			indIter++;
			//System.out.println(indIter);
		}
		
		//PHASE 4) Postprocessing
		//Filter overlapping complexes
		checkOverlap(rankAlign, overlapThresh);
        
		//Print final set of local alignments
		fileMan.writeAlignments(g, rankAlign, pathOutput);
		
		//End algorithm
		double fine=System.currentTimeMillis();
		//System.out.println("Running time: "+(((float)(fine-inizio))/1000)+" sec \n");
		
		return rankAlign;
	}
	
	//Filter the set of candidate seeds to nodes having degree above a threshold
	//@g: the set of input graphs
	//@thresh: the threshold for the minimum degree of nodes
	public HashSet<String>[] filterStartingNodes(Graph[] g, int thresh)
	{
		int i=0;
		HashSet<String>[] startingNodes=new HashSet[g.length];
		for(i=0;i<g.length;i++)
		{
			startingNodes[i]=new HashSet<String>();
			HashMap<String,Node> mapNodes=g[i].getMapNodes();
			Iterator<String> it=mapNodes.keySet().iterator(); 
			while(it.hasNext())
			{
				String id=it.next();
				int degree=mapNodes.get(id).degree();
				if(degree>=thresh)
					startingNodes[i].add(id);
			}
			//System.out.println(startingNodes[i].size());
		}
		return startingNodes;
	}
	
	//Returns the list of nodes adjacent to already aligned nodes in each network
	//@g: the set of input graphs
	//@seed: the set of already aligned nodes in each graph (array indexed by graph ids)
	public HashSet<String>[] getAdiacSeed(Graph[] g, Vector<String>[] seed)
	{
		HashSet<String>[] adiacSeed=new HashSet[g.length];
		int i=0, j=0, k=0;
		for(i=0;i<adiacSeed.length;i++)
		{
			adiacSeed[i]=new HashSet<String>();
			//Take the adjacent nodes to each aligned node in the i-th network and add them to the final set of adjacent nodes
			for(j=0;j<seed[i].size();j++)
			{
				HashMap<String,Double> adiac=g[i].getMapNodes().get(seed[i].get(j)).getAdiacs();
				Iterator<String> it=adiac.keySet().iterator();
				while(it.hasNext())
				{
					String adiacNode=it.next();
					adiacSeed[i].add(adiacNode);
				}
			}
			for(j=0;j<seed[i].size();j++)
				adiacSeed[i].remove(seed[i].get(j));
		}
		return adiacSeed;
	}
	
	//Initialize maps of labels and adjancency connection profiles for each candidate node for the extension step in each network
	//@g: the set of input graphs
	//@seed: the set of already aligned nodes in each network
	//@adiacSeed: the set of nodes adjacent to already aligned nodes in each network
	//@labels: the set of labels of nodes adjacent to already aligned nodes
	//@adiacMaps: the connection profiles of nodes adjacent to already aligned nodes. 
	//adiacMaps[i][j][k] is true iff the j-th adjacent node to aligned nodes of the i-th network is connected to the k-th aligned node of the i-th network
	public void buildMaps(Graph[] g, Vector<String>[] seed, HashSet<String>[] adiacSeed, String[][] labels, boolean[][][] adiacMaps)
	{
		int i=0, j=0, k=0;
		for(i=0;i<adiacSeed.length;i++)
		{
			labels[i]=new String[adiacSeed[i].size()];
			adiacMaps[i]=new boolean[adiacSeed[i].size()][seed[i].size()];
			Iterator<String> it=adiacSeed[i].iterator();
			j=0;
			while(it.hasNext())
			{
				String id=it.next();
				labels[i][j]=id;
				adiacMaps[i][j]=getAdiacMap(g[i],seed[i],id);
				j++;
			}
		}
	}
    
	//Returns the connection profile array of a node
	//@g: one of the input graphs
	//@seed: the set of already aligned nodes in g
	//@nodeRef: the node for which we have to compute the connection profile array
	//mapRef[i] is true iff "nodeRef" node is connected to the i-th node in the set of already aligned nodes of g
	public boolean[] getAdiacMap(Graph g, Vector<String> seed, String nodeRef)
	{
		boolean[] mapRef=new boolean[seed.size()];
		HashMap<String,Double> adiacRef=g.getMapNodes().get(nodeRef).getAdiacs();
		int k=0;
		for(k=0;k<seed.size();k++)
		{
			if(adiacRef.containsKey(seed.get(k)))
				mapRef[k]=true;
			else
				mapRef[k]=false;
		}
		return mapRef;
	}
	
	//Extend the current local alignment by adding one node at the time to each already aligned subgraph
	//@g: the set of input graphs
	//@seed: the set of already aligned nodes in each network
	//@iterGibbs: the number of iterations of Gibbs sampling
	//@structConservationThresh: minimum ISC score for the current local alignment
	public void extendSeeds(Graph[] g, Vector<String>[] seed, int iterGibbs, double structConservationThresh)
	{
		int i=0, j=0;
		boolean extend=true;
		while(extend)
		{
			//Check if extension is possible
			//Check if aligned nodes have adjacent nodes
			HashSet<String>[] adiacSeed=getAdiacSeed(g,seed);
			for(i=0;i<adiacSeed.length;i++)
			{
				if(adiacSeed[i].isEmpty())
					extend=false;
			}
			if(extend)
			{
				//Initialize data structures for Gibbs sampling
				boolean[][][] adiacMaps= new boolean[g.length][][];
				String[][] labels=new String[g.length][];
				buildMaps(g, seed, adiacSeed, labels, adiacMaps);
				for(i=0;i<labels.length;i++)
				{
					if(labels[i].length==0)
						extend=false;
				}
				if(extend)
				{
					//Run Gibbs sampling
					GibbsSampler gibbs=new GibbsSampler(iterGibbs);
					String[] alignedNodes=gibbs.runGibbs(labels,adiacMaps);
					//Extend the current alignment with last aligned nodes
					for(i=0;i<alignedNodes.length;i++)
						seed[i].add(alignedNodes[i]);
					//Compute the ISC score of current alignment
					double avgStructConserv=getScoreTopology(g,seed);
					//Check if the ISC score of the alignment is above the threshold
					if(avgStructConserv<structConservationThresh)
					{
						//Remove newly aligned nodes from the current alignment and stop extension step
						for(i=0;i<alignedNodes.length;i++)
							seed[i].removeElementAt(seed[i].size()-1);
						extend=false;
					}
				}
			}
		}
	}
	
	//Compute the topological ISC score of the local alignment
	//@g: set of input graphs
	//@seed: set of aligned nodes in each network
	public double getScoreTopology(Graph[] g, Vector<String>[] seed)
	{
		int i=0, j=0, k=0;
		double scoreTopology=0.0;
		for(i=0;i<seed.length;i++)
	        {
			for(j=i+1;j<seed.length;j++)
			{
				//Compute the topological similarity score between aligned subgraphs in the i-th and j-th network as the sum of similarities between the connection profiles of the aligned nodes
				double scorePair=0.0;
				for(k=0;k<seed[i].size();k++)
				{
					//Compute similarity between connection profiles of aligned nodes
					boolean[] mapSource=getAdiacMap(g[i],seed[i],seed[i].get(k));
					boolean[] mapDest=getAdiacMap(g[j],seed[j],seed[j].get(k));
					scorePair+=getPairwiseScore(mapSource,mapDest);
				}
				scoreTopology+=scorePair;
			}
		}
		//Normalize the topological score in [0,1]
		scoreTopology=scoreTopology/(seed[0].size()*g.length*(g.length-1)/2);
	        return scoreTopology;
	}
	
	//Compute similarity score between connection profiles of two nodes
	//@mapSource: connection profile of the first node
	//@mapDest: connection profile of the second node
	public double getPairwiseScore(boolean[] mapSource, boolean[] mapDest)
	{
		double mapScore=0.0;
		int k=0;
		for(k=0;k<mapSource.length;k++)
		{
			if(mapSource[k]==mapDest[k])
				mapScore++;
		}
		//Normalize in [0,1]
		mapScore=mapScore/mapSource.length;
		return mapScore;
	}
	
	//Remove (if possible) a set of aligned nodes from the current alignment
	//@g: set of input graphs
	//@seed: set of already aligned nodes in each network
	//@structConservationThresh: minimum ISC score that the current local alignment must have
	public void removeNode(Graph[] g, Vector<String>[] seed, double structConservationThresh)
	{
		int i=0, j=0, k=0, l=0;
		int indexRemove=-1;
		double minGoodness=1000000;
		boolean[][] articulationPoints=new boolean[seed.length][];
		//For each network, compute articulation points, i.e. nodes that disconnect the graph, if removed
		for(i=0;i<seed.length;i++)
		{
			HashMap<String,Node> mapNodes=g[i].getMapNodes();
			//Create adjacency matrix of the aligned subgraph of the i-th network
			boolean[][] edges=new boolean[seed[i].size()][seed[i].size()];
			for(j=0;j<seed[i].size();j++)
			{
				HashMap<String,Double> adiacs=mapNodes.get(seed[i].get(j)).getAdiacs();
				for(k=j+1;k<seed[i].size();k++)
				{
					if(adiacs.containsKey(seed[i].get(k)))
						edges[j][k]=edges[k][j]=true;
				}
			}
			//Get articulation points
			articulationPoints[i]=getArticulationPoints(edges);
		}
		//Compute goodness scores
		for(j=0;j<seed[0].size();j++)
		{
			boolean isRemovable=true;
			//If one of the aligning subgraphs contains an articulation point in column j, then skip the column
			for(i=0;i<seed.length;i++)
			{
				if(articulationPoints[i][j])
				{
					isRemovable=false;
					break;
				}
			}
			if(isRemovable)
			{
				//Compute goodness as the ISC score
				double goodness=0.0;
				for(i=0;i<seed.length;i++)
				{
					boolean[] mapSource=getAdiacMap(g[i],seed[i],seed[i].get(j));
					for(l=i+1;l<seed.length;l++)
					{
						boolean[] mapDest=getAdiacMap(g[l],seed[l],seed[l].get(j));
						goodness+=getPairwiseScore(mapSource,mapDest);
					}
				}
				//Add this column of aligned nodes to the candidate set for removal iff the goodness is under the ISC score threshold
				if(goodness<structConservationThresh && goodness<minGoodness)
				{
					indexRemove=j;
					minGoodness=goodness;
				}
			}
                }
		//Remove a set of aligned nodes among the candidate set (if there exist candidates)
		if(indexRemove!=-1)
		{
			for(i=0;i<seed.length;i++)
				seed[i].remove(indexRemove);
		}
	}
	
	//Filter the set of highly overlapping alignments
	//@rankAlign: list of local alignments, ordered by size and ISC score
	//@overlapThresh: minimum overlap degree between two distinct alignments
	public void checkOverlap(OrderedList<Alignment> rankAlign, double overlapThresh)
	{
		int i=0, j=0, k=0;
		//Scan alignments starting from the best one, i.e. the biggest one
		NodeOrdList<Alignment> aux=rankAlign.getMax();
		//Save observed nodes in the alignment in a Hash Set
		HashSet<String> nodiOverlap=new HashSet<String>();
		while(aux!=null)
		{
			Vector<String>[] mapping=aux.getInfo().getMapping();
			//Compute in each network the percentage of nodes that have been observed in previously analized alignments
			double[] overlapping=new double[mapping.length];
			for(i=0;i<overlapping.length;i++)
				overlapping[i]=0.0;
			for(i=0;i<mapping.length;i++)
			{
				for(j=0;j<mapping[i].size();j++)
				{
					if(nodiOverlap.contains(mapping[i].get(j)))
						overlapping[i]++;
				}
			}
			//Compute the average of this percentages across the aligning networks
			double avg=0.0;
			for(i=0;i<overlapping.length;i++)
			{
				overlapping[i]=overlapping[i]/mapping[0].size();
				avg+=overlapping[i];
			}
			avg=avg/overlapping.length;
			//Check if the average overlap degree is below the threshold, otherwise delete the alignment
			if(avg<=overlapThresh)
			{
				//Add aligned nodes to the set of observable nodes
				for(i=0;i<mapping.length;i++)
				{
					for(j=0;j<mapping[i].size();j++)
						nodiOverlap.add(mapping[i].get(j));
				}
			}
			else
				rankAlign.delete(aux.getInfo());
			aux=aux.getNext();
		}
	}
	
	//Auxiliary recursive method for computing articulation points
	public void APUtil(int u, boolean[][] edge, boolean[] visited, int[] disc, int[] low, int[] parent, boolean[] ap, int time)
	{
		int children = 0;
		visited[u]=true;
		time++;
		disc[u]=low[u]=time;
		int v=0;
		for(v=0;v<edge[u].length;v++)
		{
			if(edge[u][v])
			{
				if(!visited[v])
				{
					children++;
					parent[v]=u;
					APUtil(v, edge, visited, disc, low, parent, ap, time);
					low[u] = Math.min(low[u],low[v]);
					if(parent[u]==-1 && children>1)
						ap[u]=true;
					if(parent[u]!=-1 && low[v]>=disc[u])
						ap[u]=true;
				}
				else if(parent[u]!=-1 && parent[u]!=v)
					low[u]=Math.min(low[u], disc[v]);
			}
		}
	}
	
	//Returns the articulation points in the graph, i.e. the nodes that disconnect the graph if they are removed
	//@graph: boolean adjacency matrix of a network
	public boolean[] getArticulationPoints(boolean[][] graph)
	{
		boolean[] visited=new boolean[graph.length];
		int[] disc=new int[graph.length];
		int[] low=new int[graph.length];
		int[] parent=new int[graph.length];
		int i=0;
		for(i=0;i<parent.length;i++)
			parent[i]=-1;
		boolean[] ap=new boolean[graph.length];
		int time=0;
		for(i=0;i<graph.length;i++)
		{
			if(!visited[i])
				APUtil(i, graph, visited, disc, low, parent, ap, time);
		}
		return ap;
	}
	
}