import java.util.HashMap;
import java.util.Random;

//Class for Gibbs sampling method
//@numIter: number of iterations of Gibbs sampling
//@random: source of randomness
public class GibbsSampler
{
	private int numIter;
	private Random random;
	
	public GibbsSampler(int numIter)
	{
		this.numIter=numIter;
		this.random=new Random();
	}
	
	//runGibbs: run Gibbs sampling algorithm for local alignment
	//@labels: set of labels of candidate nodes for the alignment, in each network
	//@adiacMaps: connection profile maps of nodes with respect to the already aligned nodes in each network (used only in the extension steps)
	public String[] runGibbs(String[][] labels, boolean[][][] adiacMaps)
	{
		//Build initial alignment with random entries
		int[] alignment=new int[labels.length];
		int i=0, j=0;
		for(i=0;i<alignment.length;i++)
		{
			int row=random.nextInt(labels[i].length);
			alignment[i]=row;
		}
		int indIter=0;
		//Store the initial alignment as the best current one
		double bestScoreAlign=getScoreAlignment(alignment,labels,adiacMaps);
		int[] bestAlign=new int[alignment.length];
		for(i=0;i<alignment.length;i++)
			bestAlign[i]=alignment[i];
		for(indIter=0;indIter<numIter;indIter++)
		{
			//Select randomly a node in the current alignment
			int indexSost=random.nextInt(alignment.length);
			//Compute probabilities of replacing the removed node with a candidate node of the same network
			double[] probs =computeLR(alignment, indexSost, labels, adiacMaps);
			//Order probabilities in descending order and store the permutation representing the order
			int[] probs_ids=new int[probs.length];
			for(i=0;i<probs_ids.length;i++)
				probs_ids[i]=i;
			insertionSort(probs, probs_ids);
			//Sample a node according to the probabilities and add it to the current alignment
			alignment[indexSost]=sample(probs,probs_ids);
			//Compute the score of the alignment and update best alignment, if necessary
			double scoreAlign=getScoreAlignment(alignment,labels,adiacMaps);
			if(scoreAlign>bestScoreAlign)
			{
				bestScoreAlign=scoreAlign;
				for(i=0;i<alignment.length;i++)
					bestAlign[i]=alignment[i];
			}
		}
		//Return the final alignment
		String[] finalAlign=new String[alignment.length];
		for(i=0;i<finalAlign.length;i++)
			finalAlign[i]=labels[i][bestAlign[i]];
		return finalAlign;
	}
	
	//computeLR: compute the Likelihood Ratio, i.e. the probability of selecting a candidate node in a network
	//@alignment: current alignment
	//@index: the index of the network the removed node belongs to
	//@labels: the set of labels of candidate nodes in the index-th network
	//@adiacMaps: connection profile maps of candidate nodes in the index-th network (used only in the extension steps)
	public double[] computeLR(int[] alignment, int index, String[][] labels, boolean[][][] adiacMaps)
	{
		double[] probSelect=new double[labels[index].length];
		int i=0, j=0, k=0;
		double denom=0.0;
		for(i=0;i<labels[index].length;i++)
		{ 
			//Compute similarities between the candidate node (refNode) and the nodes in the current alignment (excluding the removed node)
			String refNode=labels[index][i];
			//Label similarity score
			double scoreHomology=1.0;
			//Structural similarity score
			double scoreAdiacs=1.0;
			for(j=0;j<alignment.length;j++)
			{
				if(j!=index)
				{
					//Compute label and structural similarity score between refNode and a node in the current alignment
					String alignNode=labels[j][alignment[j]];
					if(adiacMaps!=null)
					{
						double pairAdiacs=0.0;
						for(k=0;k<adiacMaps[index][i].length;k++)
						{
							if(adiacMaps[j][alignment[j]][k]==adiacMaps[index][i][k])
								pairAdiacs++;
						}
						//Adjust to avoid zero score
						if(pairAdiacs==0.0)
							pairAdiacs+=0.01;
						scoreAdiacs*=pairAdiacs;
					}
				}
			}
			//Combine label and topology score in a unique score
			probSelect[i]=scoreHomology*scoreAdiacs;
			denom+=probSelect[i];
		}
		//Normalize the overall scores in [0,1] to obtain probabilities
		for(i=0;i<probSelect.length;i++)
			probSelect[i]=probSelect[i]/denom;
		return probSelect;
	}
	
	//Sample a node according a non-uniform probability distribution
	//@prob: probability distribution, where probabilities are in descending order
	//@prob_ids: permutation of the ordered sequence of probabilities
	public int sample(double[] prob, int[] probs_ids)
	{
		int i=0;
		int indSample=0;
		boolean found=false;
		double rand=random.nextDouble();
		//Substract probabilities to the random number until the latter becomes negative
		while(i<prob.length && !found)
		{
			rand=rand-prob[probs_ids[i]];
			if(rand<0)
			{
				indSample=probs_ids[i];
				found=true;
			}
			i++;
		}
		return indSample;
	}
	
	//Insertion sort algorithm
	//@probs: input array of probabilities (update in place)
	//@probs_ids: permutation of the ordered sequence (update in place)
	public void insertionSort(double[] probs, int[] probs_ids)
	{
		int i=0, j=0;
		for(j=1;j<probs.length;j++)
		{
			int key=probs_ids[j];
			i=j-1;
			while(i>=0 && probs[probs_ids[i]]>probs[key])
			{
				probs_ids[i+1]=probs_ids[i];
				i=i-1;
			}
			probs_ids[i+1]=key;
		}
	}
	
	//Returns the score of the current alignment of nodes
	//@alignment: current alignment
	//@labels: set of labels of candidate nodes in each network
	//@adiacMaps: set of connection profile maps of candidate nodes in each network
	public double getScoreAlignment(int[] alignment, String[][] labels, boolean[][][] adiacMaps)
	{
		double scoreAlign=0.0;
		int i=0, j=0, k=0;
		for(i=0;i<alignment.length;i++)
		{
			String sourceNode=labels[i][alignment[i]];
			for(j=i+1;j<alignment.length;j++)
			{
				double scorePair=1.0;
				String destNode=labels[j][alignment[j]];
				//Compute pairwise label and structural similarity score between two aligned nodes
				if(adiacMaps!=null)
				{
					double pairAdiacs=0.0;
					for(k=0;k<adiacMaps[i][alignment[i]].length;k++)
					{
						if(adiacMaps[j][alignment[j]][k]==adiacMaps[i][alignment[i]][k])
							pairAdiacs++;
					}
					if(pairAdiacs==0.0)
						pairAdiacs+=0.01;
					scorePair*=pairAdiacs;
				}
				scoreAlign+=scorePair;
			}
		}
		return scoreAlign;
	}
}